package main;

import static pa.cl.OpenCL.*;

import java.nio.IntBuffer;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.util.BufferHelper;
import pa.util.IOUtil;
import pa.util.SizeOf;
import pa.util.math.MathUtil;

public class MatrixMultiplikation 
{
    
    public static void dim1(CLKernel kernel, CLCommandQueue queue, int rowsA, int colsB)
    {
        PointerBuffer gws = new PointerBuffer(1);
        gws.put(0, rowsA * colsB);
        
        clEnqueueNDRangeKernel(queue, kernel, 1, null, gws, null, null, null);
    }
    
    public static void dim2(CLKernel kernel, CLCommandQueue queue, int rowsA, int colsB)
    {
        PointerBuffer gws = new PointerBuffer(2);
        gws.put(0, rowsA);
        gws.put(1, colsB);
        
        clEnqueueNDRangeKernel(queue, kernel, 2, null, gws, null, null, null);
    }
    
    public static void main(String[] a)
    {
        int rowsA = 10;
        int colsA = 5;
        
        int rowsB = 5;
        int colsB = 10;
        
        if(colsA != rowsB)
        {
            throw new IllegalArgumentException("rowsA != colsB");
        }
        
        CLUtil.createCL();
        
        PlatformDevicePair pair = CLUtil.choosePlatformAndDevice();
        
        String source = IOUtil.readFileContent("programs/MatrixMultiplication.cl");
        
        CLContext context = clCreateContext(pair.platform, pair.device, null, null);
        
        CLCommandQueue queue = clCreateCommandQueue(context, pair.device, 0);
        
        CLProgram program = clCreateProgramWithSource(context, source);
        
        clBuildProgram(program, pair.device, "", null);

        CLKernel kernel1d = clCreateKernel(program, "matrixMul1d");        
        CLKernel kernel2d = clCreateKernel(program, "matrixMul2d");
        
        IntBuffer host_A = BufferUtils.createIntBuffer(rowsA * colsA);
        IntBuffer host_B = BufferUtils.createIntBuffer(rowsB * colsB);
        IntBuffer host_C = BufferUtils.createIntBuffer(rowsA * colsB);

        for(int i = 0; i < rowsA * colsA; ++i)
        {
            host_B.put(MathUtil.nextInt(3));
        }
        
        for(int i = 0; i < rowsB * colsB; ++i)
        {
            host_A.put(MathUtil.nextInt(3));
        }

        BufferHelper.printBuffer(host_A, colsA, "A=");

        BufferHelper.printBuffer(host_B, colsB, "\nB=");
        
        host_A.flip();
        host_B.flip();
        
        CLMem A = clCreateBuffer(context, CL_MEM_COPY_HOST_PTR | CL_MEM_READ_ONLY, host_A);
        CLMem B = clCreateBuffer(context, CL_MEM_COPY_HOST_PTR | CL_MEM_READ_ONLY, host_B);
        CLMem C = clCreateBuffer(context, CL_MEM_WRITE_ONLY, rowsA * colsB * SizeOf.FLOAT);
        
        clSetKernelArg(kernel1d, 0, A);
        clSetKernelArg(kernel1d, 1, B);
        clSetKernelArg(kernel1d, 2, C);
        clSetKernelArg(kernel1d, 3, colsA);
        clSetKernelArg(kernel1d, 4, colsB);
        
        clSetKernelArg(kernel2d, 0, A);
        clSetKernelArg(kernel2d, 1, B);
        clSetKernelArg(kernel2d, 2, C);
        clSetKernelArg(kernel2d, 3, colsA);
        clSetKernelArg(kernel2d, 4, colsB);
        
        dim1(kernel1d, queue, rowsA, colsB);
        //dim2(kernel2d, queue, rowsA, colsB);
        
        clEnqueueReadBuffer(queue, C, CL_FALSE, 0, host_C, null, null);
        
        clFinish(queue);
        
        BufferHelper.printBuffer(host_C, colsB, "\nC=(A*B)=");
        
        clReleaseCommandQueue(queue);
        clReleaseKernel(kernel1d);
        clReleaseProgram(program);
        clReleaseMemObject(A);
        clReleaseMemObject(B);
        clReleaseMemObject(C);
        clReleaseContext(context);
       
        CLUtil.destroyCL();
    }
}
