package main;

import static pa.cl.OpenCL.CL_FALSE;
import static pa.cl.OpenCL.CL_MEM_COPY_HOST_PTR;
import static pa.cl.OpenCL.CL_MEM_READ_ONLY;
import static pa.cl.OpenCL.clBuildProgram;
import static pa.cl.OpenCL.clCreateBuffer;
import static pa.cl.OpenCL.clCreateCommandQueue;
import static pa.cl.OpenCL.clCreateContext;
import static pa.cl.OpenCL.clCreateKernel;
import static pa.cl.OpenCL.clCreateProgramWithSource;
import static pa.cl.OpenCL.clEnqueueNDRangeKernel;
import static pa.cl.OpenCL.clEnqueueReadBuffer;
import static pa.cl.OpenCL.clFinish;
import static pa.cl.OpenCL.clReleaseCommandQueue;
import static pa.cl.OpenCL.clReleaseContext;
import static pa.cl.OpenCL.clReleaseKernel;
import static pa.cl.OpenCL.clReleaseMemObject;
import static pa.cl.OpenCL.clReleaseProgram;
import static pa.cl.OpenCL.clSetKernelArg;

import java.nio.IntBuffer;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.util.BufferHelper;
import pa.util.IOUtil;
import pa.util.SizeOf;

public class VectorTransformation 
{    
    public static int getIndexRowMO(int i, int j, int m)
    {
        return i * m + j;
    }   
    
    // Berechnet und schreibt Komponente i des n x 1 Vektors c, mit c = A * b.
    // Dabei ist A eine n x m Matrix und b ein m x 1 Vektor
    public static void calcMatVecProductComponent(int[] A, int[] b, int[] c, final int m, int i)
    {
        int sum = 0;
        for(int k = 0; k < m; ++k)
        {
            sum += A[getIndexRowMO(i, k, m)] * b[k];
        }
        c[i] = sum;
    } 
    
    public static void main(String[] args) 
    {
        int n = 3, m = 2;       
        int[] A_Java = new int[n * m], b_Java = new int[m], c_Java = new int[n];  
        IntBuffer A_Host = BufferUtils.createIntBuffer(n * m);
        IntBuffer b_Host = BufferUtils.createIntBuffer(m);
        IntBuffer c_Host = BufferUtils.createIntBuffer(n);
        
        A_Java[0] = 1; A_Java[1] = 2;
        A_Java[2] = 3; A_Java[3] = 4;
        A_Java[4] = 1; A_Java[5] = 0;       

        b_Java[0] = 2;
        b_Java[1] = 5;
        
        c_Java[0] = 0;
        c_Java[0] = 0;

        A_Host.put(A_Java);
        A_Host.rewind();
        b_Host.put(b_Java);
        b_Host.rewind();
        c_Host.put(c_Java);
        c_Host.rewind();

        /*
         * Berechnung mit java
         */
        for(int i = 0; i < n; ++i)
        {
            calcMatVecProductComponent(A_Java, b_Java, c_Java, m, i);
        }   
        System.out.println("Berechnung des Produkts mit Java");
        printMat(A_Java,n,m);
        System.out.println("\n" + "*" + "\n");
        printMat(b_Java,m,1);
        System.out.println("\n" + "=" + "\n");
        printMat(c_Java,n,1);
        System.out.println();
        
        /*
         * Berechnung mit OpenCL
         */
        System.out.println("Berechnung des Produkts mit OpenCL");
        CLUtil.createCL();
        
        PlatformDevicePair pair = CLUtil.choosePlatformAndDevice();
        
        String source = IOUtil.readFileContent("programs/VectorTransformation.cl");
        
        CLContext context = clCreateContext(pair.platform, pair.device, null, null);
        
        CLCommandQueue queue = clCreateCommandQueue(context, pair.device, 0);
        
        CLProgram program = clCreateProgramWithSource(context, source);
        
        clBuildProgram(program, pair.device, "", null);

        CLKernel kernel = clCreateKernel(program, "calcMatVecProductComponent");
        CLMem A = clCreateBuffer(context, CL_MEM_COPY_HOST_PTR | CL_MEM_READ_ONLY, A_Host);   
        CLMem b = clCreateBuffer(context, CL_MEM_COPY_HOST_PTR | CL_MEM_READ_ONLY, b_Host);
        CLMem c = clCreateBuffer(context, 0, c_Host.capacity() * SizeOf.INTEGER);        
        
        int indexDim = 1;
        PointerBuffer gws = new PointerBuffer(indexDim);
        gws.put(0, n);
        
        clSetKernelArg(kernel, 0, A);
        clSetKernelArg(kernel, 1, b);
        clSetKernelArg(kernel, 2, c);
        clSetKernelArg(kernel, 3, m);

        clEnqueueNDRangeKernel(queue, kernel, indexDim, null, gws, null, null, null);
        
        clEnqueueReadBuffer(queue, c, CL_FALSE, 0, c_Host, null, null);       
        clFinish(queue);
        
        System.out.println();
        printMat(A_Host,n,m);
        System.out.println("\n" + "*" + "\n");
        printMat(b_Host,m,1);
        System.out.println("\n" + "=" + "\n");
        printMat(c_Host,n,1);       

        clReleaseCommandQueue(queue);
        clReleaseKernel(kernel);
        clReleaseProgram(program);
        clReleaseMemObject(A);
        clReleaseMemObject(b);
        clReleaseMemObject(c);
        clReleaseContext(context);
       
        CLUtil.destroyCL();
    }
    
    public static void printMat(IntBuffer mat, int n, int m)
    {
        BufferHelper.printBuffer(mat, m);
    }
    
    public static void printMat(int[] mat, int n, int m)
    {
        IntBuffer _im = BufferUtils.createIntBuffer(mat.length);
        _im.put(mat);
        printMat(_im, n, m);       
    }
}
